<?php

namespace Drupal\flow\Entity;

use Drupal\flow\Flow;

/**
 * Handles the saving of multiple entities.
 */
class EntitySaveHandler {

  /**
   * The saving list threshold.
   *
   * When the number of entities in a saving list surpasses this value, then
   * saving will be performed.
   *
   * @var int
   */
  protected static int $threshold = 20;

  /**
   * Constructs a new entity save handler instance.
   *
   * @param int $threshold
   *   The saving list threshold.
   */
  public function __construct(int $threshold = 20) {
    static::$threshold = $threshold;
  }

  /**
   * Get the entity save handler service.
   *
   * @return \Drupal\flow\Entity\EntitySaveHandler
   *   The entity save handler.
   */
  public static function service(): EntitySaveHandler {
    return \Drupal::service('flow.entity_save_handler');
  }

  /**
   * Saves the given list of entities, if it makes sense to do so.
   *
   * Any entity that got saved will be removed from the given list.
   * This method should only be called when it's guaranteed that somewhere later
   * within the same PHP process ::ensureSaveAll() will be finally called.
   *
   * @param \Drupal\Core\Entity\EntityInterface[] &$entities
   *   The list of entities to save.
   */
  public function saveIfRequired(array &$entities): void {
    if (count($entities) > static::$threshold) {
      $this->ensureSaveAll($entities);
    }
  }

  /**
   * Ensures that all entities are being saved that are in the given list.
   *
   * Any entity that is ensured to be saved (or got already saved) will be
   * removed from the given list. Entities may remain in the list, when they
   * are not safe to be saved now, but instead to be saved later. No further
   * call of ::ensureSaveAll() is needed, as this handler guarantees the save.
   *
   * @param \Drupal\Core\Entity\EntityInterface[] &$entities
   *   The list of entities to save.
   */
  public function ensureSaveAll(array &$entities): void {
    /** @var \Drupal\Core\Entity\ContentEntityInterface[] $ensured */
    $ensured = [];
    // Treat entities laying in the Flow stack as already ensured for saving.
    foreach (Flow::$stack as &$stacked_entities) {
      foreach ($stacked_entities as $stacked_entity) {
        if (!in_array($stacked_entity, $ensured, TRUE)) {
          if ($uuid = $stacked_entity->uuid()) {
            $ensured[$uuid] = $stacked_entity;
          }
          else {
            $ensured[] = $stacked_entity;
          }
        }
      }
    }
    foreach ($entities as $i => $entity) {
      if (in_array($entity, $ensured, TRUE) || ($entity->uuid() && isset($ensured[$entity->uuid()]))) {
        unset($entities[$i]);
        continue;
      }
      if ($entity->isNew()) {
        // When the entity is new and being referenced from an already ensured
        // entity, then this entity would be automatically saved. If not, it
        // will be saved via insert and update hook implementation afterwards.
        /** @var \Drupal\Core\Entity\ContentEntityInterface $back */
        foreach ($ensured as $back) {
          foreach ($back->referencedEntities() as $referenced) {
            if (($referenced === $entity) || ($entity->uuid() && $entity->uuid() === $referenced->uuid())) {
              continue 3;
            }
          }
        }
      }
      foreach ($entity->referencedEntities() as $referenced) {
        if (in_array($referenced, $ensured, TRUE) || ($referenced->uuid() && isset($ensured[$referenced->uuid()]))) {
          // When the contained reference is in the list of ensured entities,
          // it is within its own saving process. For that case, this entity
          // needs to be saved afterwards. This will happen, if not otherwise
          // done, handled by flow_entity_insert() and flow_entity_update().
          continue 2;
        }
      }
      unset($entities[$i]);
      $entity->save();
      if ($uuid = $entity->uuid()) {
        $ensured[$uuid] = $entity;
      }
      else {
        $ensured[] = $entity;
      }
    }
  }

}
